@@ -2,9 +2,9 @@
 
    This file is part of the UPX executable compressor.
 
-   Copyright (C) 1996-2023 Markus Franz Xaver Johannes Oberhumer
-   Copyright (C) 1996-2023 Laszlo Molnar
-   Copyright (C) 2001-2023 John F. Reiser
+   Copyright (C) Markus Franz Xaver Johannes Oberhumer
+   Copyright (C) Laszlo Molnar
+   Copyright (C) John F. Reiser
    All Rights Reserved.
 
    UPX and the UCL library are free software; you can redistribute them
@@ -34,6 +34,8 @@
 
 #include "file.h"
 #include "filter.h"
+#define WANT_EHDR_ENUM 1
+#include "p_elf_enum.h"
 #include "linker.h"
 #include "packer.h"
 #include "p_elf.h"
@@ -74,14 +76,14 @@ PackBSDI386::PackBSDI386(InputFile *f) : super(f)
     }
 }
 
-static const
+static const CLANG_FORMAT_DUMMY_STATEMENT
 #include "stub/i386-linux.elf.execve-entry.h"
-static const
+static const CLANG_FORMAT_DUMMY_STATEMENT
 #include "stub/i386-linux.elf.execve-fold.h"
 
-static const
+static const CLANG_FORMAT_DUMMY_STATEMENT
 #include "stub/i386-bsd.elf.execve-entry.h"
-static const
+static const CLANG_FORMAT_DUMMY_STATEMENT
 #include "stub/i386-bsd.elf.execve-fold.h"
 
 
@@ -279,15 +281,6 @@ PackLinuxI386::pack4(OutputFile *fo, Filter &ft)
     fo->rewrite(&elfout, overlay_offset);
 }
 
-static unsigned
-umax(unsigned a, unsigned b)
-{
-    if (a <= b) {
-        return b;
-    }
-    return a;
-}
-
 Linker *PackLinuxI386::newLinker() const
 {
     return new ElfLinkerX86;
@@ -302,7 +295,7 @@ PackLinuxI386::buildLinuxLoader(
     Filter const *ft
 )
 {
-    initLoader(proto, szproto);
+    initLoader(EM_386, proto, szproto);
 
     unsigned fold_hdrlen = 0;
   if (0 < szfold) {
@@ -311,7 +304,7 @@ PackLinuxI386::buildLinuxLoader(
          usizeof(l_info);
     if (0 == get_le32(fold_hdrlen + fold)) {
         // inconsistent SIZEOF_HEADERS in *.lds (ld, binutils)
-        fold_hdrlen = umax(0x80, fold_hdrlen);
+        fold_hdrlen = upx::umax(0x80u, fold_hdrlen);
     }
   }
     // This adds the definition to the "library", to be used later.
@@ -379,12 +372,21 @@ PackLinuxI386::buildLinuxLoader(
             (res->lit_context_bits << 0) |
             (res->lit_pos_bits << 8) |
             (res->pos_bits << 16);
-        if (linker->bele->isBE()) // big endian - bswap32
+        if (bele->isBE()) // big endian - bswap32
             properties = bswap32(properties);
         linker->defineSymbol("lzma_properties", properties);
-        // -2 for properties
-        linker->defineSymbol("lzma_c_len", ph.c_len - 2);
-        linker->defineSymbol("lzma_u_len", ph.u_len);
+
+        // These lengths assume only one block (typ. 524288 bytes: 0.5 MiB).
+        // i386 handles more than one block, and computes the lengths
+        // dynamically from struct b_info.  Why do others need these?
+        if (linker->findSymbol("lzma_c_len", false)) {
+            // -2 for properties
+            linker->defineSymbol("lzma_c_len", ph.c_len - 2);
+        }
+        if (linker->findSymbol("lzma_u_len", false)) {
+            linker->defineSymbol("lzma_u_len", ph.c_len);
+        }
+
         unsigned const stack = getDecompressorWrkmemSize();
         linker->defineSymbol("lzma_stack_adjust", 0u - stack);
     }
@@ -446,7 +448,7 @@ int PackLinuxI386::getLoaderPrefixSize() const
 
 
 /*************************************************************************
-// some ELF utitlity functions
+// some ELF utility functions
 **************************************************************************/
 
 // basic check of a Linux ELF Ehdr
@@ -495,7 +497,7 @@ int PackLinuxI386::checkEhdr(const Elf_LE32_Ehdr *ehdr) const
 //
 **************************************************************************/
 
-bool PackLinuxI386::canPack()
+tribool PackLinuxI386::canPack()
 {
     if (exetype != 0)
         return super::canPack();
