@@ -2,9 +2,9 @@
 
    This file is part of the UPX executable compressor.
 
-   Copyright (C) 1996-2022 Markus Franz Xaver Johannes Oberhumer
-   Copyright (C) 1996-2022 Laszlo Molnar
-   Copyright (C) 2000-2022 John F. Reiser
+   Copyright (C) Markus Franz Xaver Johannes Oberhumer
+   Copyright (C) Laszlo Molnar
+   Copyright (C) John F. Reiser
    All Rights Reserved.
 
    UPX and the UCL library are free software; you can redistribute them
@@ -63,7 +63,7 @@
 // and then followed by a comma to ignore the return value.
 // The only complication is that percent and backslash
 // must be doubled in the format string, because the format
-// string is processd twice: once at compile-time by 'asm'
+// string is processed twice: once at compile-time by 'asm'
 // to produce the assembled value, and once at runtime to use it.
 #if defined(__powerpc__)  //{
 #define DPRINTF(fmt, args...) ({ \
@@ -250,6 +250,7 @@ ERR_LAB
             xi->size -= h.sz_cpr;
         }
         else { // copy literal block
+            xi->size += sizeof(h);  // xread(xi, &h, sizeof(h)) was a peek
             xread(xi, xo->buf, h.sz_cpr);
         }
         xo->buf  += h.sz_unc;
@@ -474,6 +475,18 @@ extern void bswap(void *, unsigned);
 
 typedef size_t Addr;  // this source file is used by 32-bit and 64-bit machines
 
+void my_bkpt(void const *foo, ...)
+{
+    (void)foo;
+#if defined(__x86_64__)  //{
+    __asm__(".byte 0xcc");
+#elif  defined(__aarch64__)  //}{
+     __asm__("brk #0");
+#elif  defined(__arm__)  //}{
+     __asm__(".long 0xe7f001f0"); // "bkpt" ==> "udf #16"
+#endif  //}
+}
+
 // Find convex hull of PT_LOAD (the minimal interval which covers all PT_LOAD),
 // and mmap that much, to be sure that a kernel using exec-shield-randomize
 // won't place the first piece in a way that leaves no room for the rest.
@@ -558,8 +571,8 @@ do_xmap(
     for ( j=0; j < mhdr->ncmds; ++j,
         (sc = (Mach_segment_command *)((sc->cmdsize>>2) + (unsigned *)sc))
     ) {
-        DPRINTF("  #%%d  cmd=%%x  cmdsize=%%x  vmsize=%%x\\n",
-                j, sc->cmd, sc->cmdsize, sc->vmsize);
+        DPRINTF("  #%%d  sc=%%x  cmd=%%x  cmdsize=%%x  vmsize=%%x\\n",
+                j, sc, sc->cmd, sc->cmdsize, sc->vmsize);
         if (LC_SEGMENT==sc->cmd) {
             struct b_info h;
             if (xi && sc->filesize) { // Find the correct compressed block.
@@ -712,7 +725,6 @@ do_xmap(
     return rv;
 }
 
-
 /*************************************************************************
 // upx_main - called by our unfolded entry code
 //
@@ -764,6 +776,7 @@ upx_main(
             err_exit(18);
         }
 fat:
+        DPRINTF("fat_offset= %%p\\n", fat_offset);
         if ((ssize_t)sz_mhdr!=pread(fdi, (void *)mhdr, sz_mhdr, fat_offset)) {
 ERR_LAB
             err_exit(19);
